# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Derive from Singleton to create a new singleton.
"""


class Singleton(type):
    """Create a singleton by using this as a metaclass.
    Example1: class MySingleton(metaclass=Singleton):
    Example2: class MySingleton(MyBaseClass, metaclass=Singleton):
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            # Calls __init__ on new/only instance.
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        else:
            # Calls __init__ on existing/only instance (can be used to ensure same parameters or raise etc.).
            cls._instances[cls].__init__(*args, **kwargs)
        return cls._instances[cls]



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

import os
from pathlib import Path
import logging
from pytest import LogCaptureFixture, CaptureFixture, fail, raises
from atbu.common.aes_cbc import (
    AES_CBC_Base,
    AesCbcPaddingEncryptor,
    AesCbcPaddingDecryptor,
)

LOGGER = logging.getLogger(__name__)


def setup_module(module):
    pass


def teardown_module(module):
    pass


def test_encryption_decryption_simple(tmp_path: Path):
    key = bytes(os.urandom(int(256 / 8)))
    iv = bytes(os.urandom(AES_CBC_Base.BLOCK_SIZE))
    LOGGER.debug(f"key={key.hex(' ')}")
    LOGGER.debug(f"IV={iv.hex(' ')}")

    for i in range(0, AES_CBC_Base.BLOCK_SIZE * 16):
        pt = bytes([0x55] * i)
        enc = AesCbcPaddingEncryptor(key=key, IV=iv)
        assert not enc.is_finalized
        ct = enc.update(pt)
        assert not enc.is_finalized
        ct += enc.finalize()
        assert enc.is_finalized

        dec = AesCbcPaddingDecryptor(key=key, IV=iv)
        assert not dec.is_finalized
        pt2 = dec.update(ct)
        assert not dec.is_finalized
        pt2 += dec.finalize()
        assert dec.is_finalized

        LOGGER.debug(f"Count={i}:")
        LOGGER.debug(f"    ct={ct.hex(' ')}")
        LOGGER.debug(f"    pt={pt.hex(' ')}")
        LOGGER.debug(f"    pt={pt2.hex(' ')}")

        assert pt == pt2


def test_encryption_decryption_edge(tmp_path: Path):
    key = bytes(os.urandom(int(256 / 8)))
    iv = bytes(os.urandom(AES_CBC_Base.BLOCK_SIZE))
    LOGGER.debug(f"key={key.hex(' ')}")
    LOGGER.debug(f"IV={iv.hex(' ')}")

    for i in range(0, AES_CBC_Base.BLOCK_SIZE * 16):
        pt = bytes([0x55] * i)
        enc = AesCbcPaddingEncryptor(key=key, IV=iv)
        assert not enc.is_finalized
        ct = enc.update(pt)
        assert not enc.is_finalized
        ct += enc.finalize()
        assert enc.is_finalized

        NUM_DECRYPTORS = AES_CBC_Base.BLOCK_SIZE * 3
        for d in range(NUM_DECRYPTORS):
            dec = AesCbcPaddingDecryptor(key=key, IV=iv)
            assert not dec.is_finalized
            pt2 = dec.update(ct[:d])
            assert not dec.is_finalized
            pt2 += dec.update(ct[d:])
            assert not dec.is_finalized
            pt2 += dec.finalize()
            assert dec.is_finalized
            LOGGER.debug(f"Count={i} Edge={d}:")
            LOGGER.debug(f"    ct={ct.hex(' ')}")
            LOGGER.debug(f"    pt={pt.hex(' ')}")
            LOGGER.debug(f"    pt={pt2.hex(' ')}")
            assert pt == pt2



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU common exceptions.
"""

# pylint: disable=missing-class-docstring

from dataclasses import dataclass
import sys
import traceback
from typing import Literal

ANOMALY_KINDS = Literal["exception", "cancelled", "unexpected state"]
ANOMALY_KIND_EXCEPTION = "exception"
ANOMALY_KIND_CANCELLED = "cancelled"
ANOMALY_KIND_UNEXPECTED_STATE = "unexpected state"


@dataclass
class Anomaly:
    kind: ANOMALY_KINDS
    exception: Exception = None
    message: str = None


class AtbuException(Exception):
    def __init__(self, message: str, cause=None):
        super().__init__(message)
        self.message = message
        self._cause = cause

    @property
    def cause(self):
        return self._cause

    def __str__(self):
        if not self._cause:
            return f"{self.message}"
        else:
            return f"{self.message} cause={self._cause}"


class LocationDoesNotExistException(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ConfigFileNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidConfigurationValue(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidConfigurationFile(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidConfiguration(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ConfigSectionNotFound(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ConfigValueNotFound(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ObjectDoesNotExistError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ContainerAlreadyExistsError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidContainerNameError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class ContainerAutoCreationFailed(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class RetryLimitReached(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class FileChangedWhileCalculatingHash(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class ExpectedDirectoryIsFileError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class SingletonAlreadyCreated(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class QueueListenerAlreadyStarted(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class QueueListenerNotStarted(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class UnexpectedOperation(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class EncryptionAlreadyEnabledError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class InvalidStateError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class UserAbortException(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class StateChangeDisallowedError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class SizeMistmatchError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class DateTimeMistmatchError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class DigestMistmatchError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CompareBytesMistmatchError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class AlreadyFinalizedError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class AlreadyUsedError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class EncryptionDecryptionFailure(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class StateNotYetKnownError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class InvalidCommandLineArgument(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class InvalidFunctionArgument(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class InvalidBase64StringError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


def exc_to_string(ex: Exception):
    return f"ex={ex} details: {traceback.format_exception(ex, ex, ex.__traceback__)}"


def exc_to_string_with_newlines(ex: Exception):
    traceback_list = traceback.format_exception(ex, ex, ex.__traceback__)
    traceback_str = "".join(traceback_list)
    return f"ex={ex} details: {traceback_str}"


def cur_exc_to_string():
    ex = sys.exc_info()[1]
    if ex is None:
        return "ex=<NoneType>"
    return exc_to_string(ex)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Support json encoding/decoding of multiple classes via use of single
instance of MultiEncoderDecoder.
"""
from typing import Callable
import json
import dataclasses


class MultiEncoderDecoder:
    """An instance of MultiEncoderDecoder can json encode/decode one or more
    classes. Call add_def() to inform this instance of each such class.
    """

    # pylint: disable=missing-class-docstring
    # pylint: disable=protected-access

    class Definition:
        def __init__(
            self,
            class_type: type,
            to_dict_method: Callable[[object], dict],
            from_dict_method: Callable[[object, object], None],
            constructor_arg_names: list[str],
        ):
            self.class_type = class_type
            self.to_dict_method = to_dict_method
            self.from_dict_method = from_dict_method
            self.constructor_arg_names = constructor_arg_names

    def __init__(self):
        self.enc_dec_defs = list()

    def add_def(
        self,
        class_type: type,
        to_dict_method: Callable[[object], dict],
        from_dict_method: Callable[[object, object], None],
        constructor_arg_names: list[str],
    ):
        """Inform this instance to include the specified class
        in its encoding/decoding efforts.
        """
        self.enc_dec_defs.append(
            MultiEncoderDecoder.Definition(
                class_type=class_type,
                to_dict_method=to_dict_method,
                from_dict_method=from_dict_method,
                constructor_arg_names=constructor_arg_names,
            )
        )

    def get_json_encoder_class(self):
        """Return a json.JSONEncoder to use is json.* calls
        accepting an encoder.
        """

        class CustomEncoder(json.JSONEncoder):
            owner = self

            def default(self, o):
                d = CustomEncoder.owner._default(o) 
                if not d:
                    return json.JSONEncoder.default(self, o)
                return d

        return CustomEncoder

    def get_json_decoder_class(self):
        """Returns a json.JSONDecoder to use in json.* calls
        accepting a decoder.
        """

        # pylint: disable=missing-class-docstring
        # pylint: disable=protected-access
        # pylint: disable=method-hidden

        class CustomDecoder(json.JSONDecoder):
            owner = self

            def __init__(self, *args, **kwargs):
                json.JSONDecoder.__init__(
                    self, object_hook=self._obj_hook_func, *args, **kwargs
                )

            def _obj_hook_func(self, obj):
                return CustomDecoder.owner._owner_obj_hook_func(
                    obj
                )

        return CustomDecoder

    def _default(self, obj):
        enc_dec_def: MultiEncoderDecoder.Definition
        for enc_dec_def in self.enc_dec_defs:
            if isinstance(obj, enc_dec_def.class_type):
                v = enc_dec_def.to_dict_method(obj)
                if isinstance(v, dict):
                    v["_type"] = enc_dec_def.class_type.__name__
                return v
        return None

    def _owner_obj_hook_func(self, obj):
        if "_type" not in obj:
            return obj
        obj_type_str = obj["_type"]
        enc_dec_def: MultiEncoderDecoder.Definition
        for enc_dec_def in self.enc_dec_defs:
            if enc_dec_def.class_type.__name__ == obj_type_str:
                args = [obj[k] for k in enc_dec_def.constructor_arg_names]
                o = enc_dec_def.class_type(*args)
                enc_dec_def.from_dict_method(o, obj)
                return o
        return obj


def create_dataclass_json_encoder(data_cls, is_strict: bool = True) -> json.JSONEncoder:
    """Create a json.JSONEncoder to handle serialization of a dataclass data_cls.
    This is useful for simple cases where a single dataclass is serialized as
    a single instance, or multiple instances in an list.

    Args:
        data_cls (dataclass): A user-defined dataclass.
        is_strict (bool, optional): If True, an a TypeError is raised if the encoder
            is asked to encode anything other than data_cls. Defaults to True.

    Raises:
        ValueError: If the specified data_cls is not a dataclass.
        TypeError: If, during encoding, the object tp encode is not an instance of data_cls.

    Returns:
        json.JSONEncoder: The encoder to encode the data_cls instances.
    """
    if not dataclasses.is_dataclass(data_cls):
        raise ValueError(f"Not a dataclass: cls={data_cls}")
    class Encoder(json.JSONEncoder): # pylint: disable=missing-class-docstring
        def default(self, o):
            if not isinstance(o, data_cls):
                if is_strict:
                    raise TypeError(
                        f"cannot identify dataclass for encoding: "
                        f"The obj ({o}) does not appear to be data_cls ({data_cls})"
                    )
                return json.JSONEncoder.default(self, o)
            obj_dict = { "_type" : data_cls.__name__}
            obj_dict.update(dataclasses.asdict(o))
            return obj_dict
    return Encoder


def create_dataclass_json_decoder(data_cls, is_strict: bool = True) -> json.JSONDecoder:
    """Create a json.JSONDecoder to handle deserialization of serialized data_cls
    instances.

    Args:
        data_cls (dataclass): A user-defined dataclass.
        is_strict (bool, optional): If True, an a TypeError is raised if the encoder
            is asked to decode anything other than data_cls. Defaults to True.

    Raises:
        ValueError: If the specified data_cls is not a dataclass.
        TypeError: If, during decoding, the object to decode is not an instance of data_cls.

    Returns:
        json.JSONDecoder: The decoder to decode the data_cls instances.
    """

    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring

    if not dataclasses.is_dataclass(data_cls):
        raise ValueError(f"Not a dataclass: cls={data_cls}")
    class Decoder(json.JSONDecoder):
        def __init__(self, *args, **kwargs):
            json.JSONDecoder.__init__(self, object_hook=self.obj_hook_func, *args, **kwargs)
        def obj_hook_func(self, obj):
            if "_type" not in obj or obj["_type"] != data_cls.__name__:
                if is_strict:
                    raise TypeError(
                        f"cannot identify dataclass for deocding: "
                        f"The dict ({obj}) does not appear to be data_cls ({data_cls})"
                    )
                return obj
            args = {}
            for field in dataclasses.fields(data_cls):
                if field.name in obj:
                    args[field.name] = obj[field.name]
            return data_cls(**args)
    return Decoder



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Utility/helper functions.
"""

# pylint: disable=missing-class-docstring

import base64
from enum import Enum
import fnmatch
from io import SEEK_END, SEEK_SET
import io
import os
from pathlib import Path
import platform
import random
import re
from shutil import copy2
from typing import Iterator, Union
from datetime import datetime, timezone

from .exception import InvalidFunctionArgument


class AutoNameEnum(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name


def is_platform_path_case_sensitive():
    return os.path.normcase("A") == "A"


def get_trash_bin_name():
    if platform.system() == "Windows":
        return "Recycle Bin"
    else:
        return "Trash"


def get_file_max_pos(fd):
    cur_pos = fd.tell()
    fd.seek(0, SEEK_END)
    end_pos = fd.tell()
    fd.seek(cur_pos, SEEK_SET)
    return end_pos


def prompt_YN(
    prompt_msg: str,
    prompt_question: str = None,
    default_enter_ans: str = "y",
    choices=None,
) -> str:
    """Return 'y' or 'n' depending on the answer entered."""
    if prompt_question is None:
        prompt_question = "Do you want to continue?"
    default_enter_ans = default_enter_ans.lower()
    if choices is None:
        choices = {"y": "[Y/n]", "n": "[y/N]"}
    if default_enter_ans not in choices.keys():
        raise ValueError(
            f"The default_enter_ans of '{default_enter_ans}' is not a possible choice of {choices.keys()}."
        )
    YN_str_to_use = choices.get(default_enter_ans)
    if YN_str_to_use is None:
        raise ValueError(
            f"The default_enter_ans is '{default_enter_ans}' but must be one of {choices.keys()}"
        )
    a = None
    while a not in choices.keys():
        if a is not None:
            print(f"Invalid response: {a}")
        print(prompt_msg)
        a = input(f"{prompt_question} {YN_str_to_use} ").lower()
        if a == "":
            a = default_enter_ans
    return a.lower()[0]


def menu_prompt(menu: dict, prompt: str, random_order: bool = False):
    if not isinstance(menu, dict):
        raise ValueError("The menu argument must be a dict.")
    order = sorted(menu.keys())
    while True:
        if random_order:
            random.shuffle(order)
        for k in order:
            print(f"{k}) {menu[k]['name']}")
        a = input(prompt).lower()
        if a not in menu.keys():
            print("Invalid choice, try again.")
            continue
        break
    func = menu[a].get("func")
    if func:
        return func()
    return a


def deep_union_dicts(dest: dict, src: dict):
    for src_k, src_v in src.items():
        dest_v = dest.get(src_k)
        if isinstance(dest_v, dict) and isinstance(src_v, dict):
            deep_union_dicts(dest_v, src_v)
        else:
            # Overwrite existing d1 value.
            dest[src_k] = src_v
    return dest


def get_path_sep_list():
    return [os.path.sep] if os.path.altsep is None else [os.path.sep, os.path.altsep]


def get_path_sep_re():
    if os.path.altsep is None:
        path_sep_re_group_pattern = rf"({re.escape(os.path.sep)})"
    else:
        path_sep_re_group_pattern = rf"({re.escape(os.path.sep)}|{re.escape(os.path.altsep)})"
    return re.compile(path_sep_re_group_pattern)


_PATH_SEP_RE = get_path_sep_re()


def strip_leading_sep(path: str) -> str:
    if path and path[0] in [os.sep, os.altsep]:
        path = path[1:]
    return path


def rel_path(top_level_dir: str, path: str):
    common_path = os.path.commonpath([os.path.normcase(top_level_dir), os.path.normcase(path)])
    if common_path != os.path.normcase(top_level_dir):
        raise ValueError(
            f"The path must be a subdirectory of top_level_dir: "
            f"top_level_dir={top_level_dir} path={path}"
        )
    rpath = path[len(top_level_dir) :]
    return strip_leading_sep(rpath)


def strip_common_path(path1: str, path2: str) -> tuple[str, str]:
    common_path = os.path.normcase(os.path.commonpath([path1, path2]))
    if not common_path:
        raise ValueError("Paths have no common path.")
    path1 = strip_leading_sep(path1[len(common_path):])
    path2 = strip_leading_sep(path2[len(common_path):])
    return path1, path2


def get_subdir_distance(path1: str, path2: str) -> int:
    if len(path1)==0 and len(path2)==0:
        return 0
    try:
        path1, path2 = strip_common_path(path1, path2)
    except ValueError:
        if path1 and path2:
            return -1
    if len(path1) >= len(path2):
        subdir_part = path1[len(path2):]
    else:
        subdir_part = path2[len(path1):]
    subdir_part = strip_leading_sep(subdir_part)
    if not subdir_part:
        return 0
    num_seps = len(_PATH_SEP_RE.findall(subdir_part))
    return num_seps + 1


def convert_to_pathlib_path(p):
    if p is None:
        return None
    if not isinstance(p, Path):
        p = Path(p)
    return p


def convert_to_str_path(p):
    if p is None:
        return p
    if not isinstance(p, str):
        p = str(p)
    return p


def pathlib_splitdrive(path: Union[str, Path]):
    if not isinstance(path, (str, Path)):
        raise ValueError(f"pathlib_splitdrive: path must be str or Path: {type(path)}")
    path_drive, path_wo_drive = os.path.splitdrive(str(path))
    return Path(path_drive), Path(path_wo_drive)


def is_absolute_path(path_to_dir: Union[str, Path]):
    if isinstance(path_to_dir, Path):
        path_to_dir = str(path_to_dir)
    converted_to_abs = os.path.normcase(os.path.abspath(path_to_dir).rstrip("\\/"))
    original_path = os.path.normcase(path_to_dir.rstrip("\\/"))
    return converted_to_abs == original_path

def iwalk_fnmatch(
    root_no_wildcards: str,
    fnmatch_pat: str
) -> Iterator[str]:
    """Walk `root_no_wildcards` and return all files matching `fnmatch_pat`.

    Args:
        root_no_wildcards (str): The root directory to search. This should not
            contain any wildcards/patterns.
        fnmatch_pat (str): An `fnmatch_path` style pattern. If this pattern
            contains "**" the search is recursive, otherwise it searches only
            the `root_no_wildcards` specified.

    Yields:
        Iterator[str]: An iterator of discovered file paths.
    """
    is_recursive = fnmatch_pat.find("**") != -1
    fnmatch_pat = os.path.normcase(fnmatch_pat)
    for root, _, files in os.walk(root_no_wildcards):
        for file in files:
            full_path = os.path.join(root, file)
            if not fnmatch.fnmatchcase(name=os.path.normcase(full_path), pat=fnmatch_pat):
                continue
            yield full_path
        if not is_recursive:
            break

def walk_fnmatch(
    root_no_wildcards: str,
    fnmatch_pat: str
) -> list[str]:
    """Walk `root_no_wildcards` and return all files matching `fnmatch_pat`.

    Args:
        root_no_wildcards (str): The root directory to search. This should not
            contain any wildcards/patterns.
        fnmatch_pat (str): An `fnmatch_path` style pattern.

    Returns:
        list[str]: A list of discovered file paths.
    """
    return list(iwalk_fnmatch(root_no_wildcards, fnmatch_pat))

def truncate_posix_timestamp(posix_timestamp):
    """Return a POSIX timestamp with fractional portion that can be represented
    by Python datetime.
    """
    return datetime.fromtimestamp(posix_timestamp, tz=timezone.utc).timestamp()


def posix_timestamp_to_datetime_utc(posix_timestamp):
    return datetime.fromtimestamp(posix_timestamp, tz=timezone.utc)


def posix_timestamp_to_ISO8601_utc_stamp(posix_timestamp):
    return posix_timestamp_to_datetime_utc(posix_timestamp=posix_timestamp).isoformat(
        timespec="microseconds"
    )


def create_numbered_backup_of_file(path: Union[str, Path], not_exist_ok: bool = False):
    path = convert_to_pathlib_path(path)
    for i in range(100000):
        candidate = path.with_name(path.name + f".{i:0>3}")
        if not candidate.exists():
            break
    if not path.exists():
        if not_exist_ok:
            return
        raise FileNotFoundError(
            f"create_numbered_backup_of_file: The source file does not exist: {str(path)}"
        )
    copy2(src=str(path), dst=str(candidate))
    return candidate


def is_valid_base64_string(str_to_check):
    if not isinstance(str_to_check, (str, bytes, bytearray)):
        raise ValueError(f"Expected str_to_check to be a str, bytes, or bytearray.")
    try:
        decoded = base64.b64decode(str_to_check)
        encoded = base64.b64encode(decoded)
        if isinstance(str_to_check, str):
            return encoded == str_to_check.encode("utf-8")
        else:
            return encoded == str_to_check
    except Exception:
        return False


def clear_file(fileobj: Union[io.IOBase, str, Path], byte_pattern: list[int] = None):
    """This should not be considered secure as it
    cannot guarantee filesystem buffering anomalies.
    """
    if isinstance(fileobj, str):
        fileobj = Path(fileobj)
    if isinstance(fileobj, Path):
        fileobj = fileobj.open("r+b")
    if not isinstance(fileobj, io.IOBase):
        raise InvalidFunctionArgument(
            f"Expecting io.IOBase derived fileobj, str path, or Path object."
        )
    if byte_pattern is None:
        byte_pattern = [0xCC, 0x55, 0x00]
    fileobj.seek(0, SEEK_END)
    total_bytes = fileobj.tell()
    for v in byte_pattern:
        b = bytes([v] * (1024 * 1024 * 25))
        num_remaining = total_bytes
        fileobj.seek(0, SEEK_SET)
        while num_remaining > 0:
            num_to_write = min(num_remaining, len(b))
            fileobj.write(b[:num_to_write])
            num_remaining -= num_to_write
        fileobj.flush()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Python profile-related helpers.
"""

import io
import logging
import os
from pathlib import Path
from .exception import (
    InvalidFunctionArgument,
    InvalidStateError,
)

class EasyProfile:
    def __init__(self, log_stats: bool = True, profile_file: str = None) -> None:
        """An EasyProfiler manages starting/stopping a Python cProfile. Upon
        calling EasyProfile.stop, the profile stats will be either dumped to
        file and/or printed to the log file as an info message.

        Either log_stats must be True and/or profile_file must be a path to a file
        which does not already exist.

        Args:
            log_stats (bool, optional): If True, print profile stats to the log
                file upon calling the stop method. Defaults to True.
            profile_file (str, optional): If not None, dump the profile stats to
                a file which can be used later with pstats. Defaults to None.

        Raises:
            InvalidFunctionArgument: If the profile_file is specified and already exists.
        """
        if profile_file is not None:
            if not isinstance(profile_file, (str, Path)) or os.path.exists(profile_file):
                raise InvalidFunctionArgument(
                    f"The profile_file must not already exist: "
                    f"profile_file={profile_file}"
                )
        self.profile_file = profile_file
        self.log_stats = log_stats
        if not self.log_stats and self.profile_file is None:
            raise (
                f"Either one or both of log_stats and/or profile_file must be specified."
            )
        self.profiler = None
        self.is_used = False

    def start(self):
        if self.is_used:
            raise InvalidStateError(f"This {EasyProfile.__name__} has already been used.")
        # pylint: disable=import-outside-toplevel
        import cProfile
        logging.info(f"Activating profiler.")
        self.profiler = cProfile.Profile()
        self.profiler.enable()
        self.is_used = True

    def stop(self):
        # pylint: disable=import-outside-toplevel
        try:
            if self.profiler is None:
                if self.is_used:
                    raise InvalidStateError(f"This profiler has already been stopped.")
                raise InvalidStateError(
                    f"This profiler has not been started yet or there was an error starting it."
                )
            logging.info(f"Disabling profiler.")
            self.profiler.disable()
            if self.profile_file is not None:
                logging.info(f"Writing profiler stats to {self.profile_file}")
                self.profiler.dump_stats(self.profile_file)
            if self.log_stats:
                import pstats
                logging.info(f"Printing profiler stats to the log.")
                s = io.StringIO()
                ps = pstats.Stats(self.profiler, stream=s).sort_stats(pstats.SortKey.CUMULATIVE)
                ps.print_stats()
                logging.info(s.getvalue())
        finally:
            self.profiler = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False
